/**
 * @file integrator.hpp
 * @brief Interfaces for integrators and systems.
 *
 * @author Kristin Siu (kasiu)
 * @author Eric Butler (edbutler)
 */

#ifndef _462_GAME_INTEGRATOR_HPP_
#define _462_GAME_INTEGRATOR_HPP_

#include "math/math.hpp"
#include <vector>

/*
   This file contains interfaces for integrators and systems, as well
   as definitions for some integrators.

   Use of this interface is optional, but recommended, as it allows you
   to easily change which integrator the system uses, and this interface
   can support arbitrary integrators. You may, however, modify or replace
   this if you wish.
 */

namespace _462 {

/**
 * Interface for an ODE system that can be solved with an integrator.
 * Similar to the interface discussed in the class notes, see those
 * and the writeup for further details.
 *
 * And integrable system is represented as a state vector of reals and the
 * current time. It provides functionality to get/set this state and compute
 * the derivative at the current time.
 *
 * The flock will be a particle system that implements this interface., where
 * the state vector is the positions and velocities of all the particles.
 */
class IntegrableSystem
{
public:

    virtual ~IntegrableSystem() { }

    /**
     * @return The number of elements in the state array.
     *   This should be constant.
     */
    virtual size_t size() const = 0;

    /**
     * Queries the current state.
     * @param arr[out] An array of at least this->size() in length. Where the
     *   state vector will be stored.
     * @param time[out] Where the current time will be stored.
     */
    virtual void get_state( real_t* arr, real_t* time ) const = 0;

    /**
     * Sets the current state, overriding the given state.
     * @param arr[in] An array of at least this->size() in length. The state
     *   vector to be set.
     * @param time[in] The time to be set.
     */
    virtual void set_state( const real_t* arr, const real_t time ) = 0;

    /**
     * Compute the derivative of the current state at the current time,
     * storing the result in the given vector.
     * @param deriv_result[out] An array of at least this->size() in length.
     *   The computed derivative of the state vector is stored here.
     */
    virtual void eval_deriv( real_t* deriv_result ) = 0;

};

/**
 * Interface for integrators that can step the simulation of a system.
 * Integrators are given the current state and must advance the simulation
 * by a given time step.
 *
 * You should, at a minimum, implement a basic Euler integrator, though
 * we suggest trying to implement more stable, higher order integrators such as
 * Midpoint (Runge-Kutta 2) or Runge-Kutta 4.
 */
class Integrator
{
public:
    virtual ~Integrator() { }

    /**
     * Step the simulation of the given system by the given timestep.
     * @param sys The system to integrate. It should be integrated starting
     *   from the system's current step.
     * @param dt The length of the time step to integrate.
     */
    virtual void integrate( IntegrableSystem& sys, real_t dt ) const = 0;

    /// meant to be used for storing state vectors locally
    /// without allocating memory every time.
    typedef std::vector< real_t > StateList;
};

/**
 * A completely incorrect integrator to demonstrate how these are used. It
 * doesn't do anything realistic at all, so don't use this in your project.
 */
class StupidIntegrator : public Integrator
{
public:
    virtual ~StupidIntegrator() { }
    virtual void integrate( IntegrableSystem& sys, real_t dt ) const;
private:
    mutable StateList state; // to avoid allocating memory every update
};

/**
 * A shell of an euler integrator for you to fill in.
 * Uses the basic Euler integration method, x' = x + dx/dt * dt.
 */
class EulerIntegrator : public Integrator
{
public:
    virtual ~EulerIntegrator() { }
    virtual void integrate( IntegrableSystem& sys, real_t dt ) const;
};

} /* _462 */

#endif /* _462_GAME_INTEGRATOR_HPP_ */

