
#include "application/application.hpp"
#include "math/camera.hpp"
#include "game/flock.hpp"

namespace _462 {

#define KEY_SCREENSHOT SDLK_f
#define KEY_LEFT SDLK_a
#define KEY_RIGHT SDLK_d
#define KEY_UP SDLK_w
#define KEY_DOWN SDLK_s
#define KEY_ZOOM_IN SDLK_u
#define KEY_ZOOM_OUT SDLK_i
#define KEY_BEND_DOWN SDLK_j
#define KEY_BEND_UP SDLK_k
#define GOAL_SPEED 30.0
#define ZOOM_SPEED 20.0
#define BEND_SPEED 1.5
#define EPSILON 1.0e-5

static const int WIDTH = 1024;
static const int HEIGHT = 768;
static const real_t FPS = 30.0;
static const char* TITLE = "15462 Project - Game";

class GameApplication : public Application
{
public:

    GameApplication() { }
    virtual ~GameApplication() { }

    virtual bool initialize();
    virtual void destroy();
    virtual void update( real_t dt );
    virtual void render();
    virtual void handle_event( const SDL_Event& event );

    // the camera
    Camera camera;

    // the flocking simulation
    Flock flock;

    // the integrator (can easily change this)
    StupidIntegrator integrator;

    // current movement of the goal
    real_t goal_x_dir; // -1, 0, 1
    real_t goal_y_dir; // -1, 0, 1

    // the position of the goal for seeking
    Vector3 goal;
};

bool GameApplication::initialize()
{
    // TODO initialization code
    glClearColor( 0.5f, 0.5f, 0.5f, 1.0f );

    // initialize goal
    goal_x_dir = 0.0;
    goal_y_dir = 0.0;
    goal = Vector3( 0.0, 0.0, 0.0 );

    glEnable( GL_DEPTH_TEST );

    // TODO change this to whatever you need
    flock.initialize( 3, goal );
    flock.set_boid( 0, Vector3( -15.0, -2.0, -1.0 ), Vector3::Zero, 1.0 );
    flock.set_boid( 1, Vector3( -13.0, 1.5, -0.4 ), Vector3::Zero, 1.0 );
    flock.set_boid( 2, Vector3( -14.0, 0.0, 0.6 ), Vector3::Zero, 1.0 );

    return true;
}

void GameApplication::destroy()
{
    // TODO cleanup code
}

void GameApplication::update( real_t dt )
{
    // TODO update code, step the simulation and apply any
    // other update code
    // move the goal
    Vector3 dir( goal_x_dir, goal_y_dir, 0.0 );
    if ( squared_length( dir ) > EPSILON ) {
        goal += ( dt * GOAL_SPEED ) * normalize( dir );
    }

    // set the goal so the flock can apply goal-seeking
    flock.set_goal( goal );

    // number of times to iterate the integrator per time step. increase this
    // if there are stability problems (or write a higher-order integrator!).
    static const size_t NUM_ITER = 20;

    // step the simulation
    real_t step_size = dt / NUM_ITER;
    for ( size_t i = 0; i < NUM_ITER; ++i ) {
        integrator.integrate( flock, step_size );
    }
}

void GameApplication::render()
{
    // query window dimensions to compute aspect ratio
    int width, height;
    this->get_dimension( &width, &height );
    real_t aspect = real_t( width ) / real_t( height );

    // TODO render code
    glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

    // TODO crappy test rendering, replace with something interesting

    // TODO make a real camera with perspective projection
    glMatrixMode( GL_PROJECTION );
    glLoadIdentity();
    static const real_t pos = 40.0;
    glOrtho( -pos * aspect, pos * aspect, -pos, pos, -100.0, 100.0 );

    glMatrixMode( GL_MODELVIEW );
    glLoadIdentity();

    // render goal
    glPointSize( 5.0 );
    glBegin( GL_POINTS );
    glColor3d( 0.0, 1.0, 0.0 );
    glVertex3d( goal.x, goal.y, goal.z );
    glEnd();

    // render flock
    flock.render();
}

void GameApplication::handle_event( const SDL_Event& event )
{
    // TODO handle imput events

    // sample event handling code, takes a screenshot when the correct key is pressed
    // you can add more event handlers in this switch as well
    switch ( event.type )
    {
    case SDL_KEYDOWN:
        switch ( event.key.keysym.sym )
        {
        case KEY_SCREENSHOT:
            take_screenshot();
            break;
        case KEY_LEFT:
            goal_x_dir = -1.0;
            break;
        case KEY_RIGHT:
            goal_x_dir = 1.0;
            break;
        case KEY_DOWN:
            goal_y_dir = -1.0;
            break;
        case KEY_UP:
            goal_y_dir = 1.0;
            break;
        default:
            break;
        }
        break;
    case SDL_KEYUP:
        switch ( event.key.keysym.sym )
        {
        case KEY_LEFT:
            if ( goal_x_dir < 0.0 )
                goal_x_dir = 0.0;
            break;
        case KEY_RIGHT:
            if ( goal_x_dir > 0.0 )
                goal_x_dir = 0.0;
            break;
        case KEY_DOWN:
            if ( goal_y_dir < 0.0 )
                goal_y_dir = 0.0;
            break;
        case KEY_UP:
            if ( goal_y_dir > 0.0 )
                goal_y_dir = 0.0;
            break;
        default:
            break;
        }
        break;
    default:
        break;
    }
}

} /* _462 */

using namespace _462;

int main( int argc, char* argv[] )
{
    GameApplication app;

    // start a new application
    return Application::start_application( &app, WIDTH, HEIGHT, FPS, TITLE );
}

